
/*
 **********************************************************************
 *     cardwo.c - PCM output HAL for emu10k1 driver
 *     Copyright 1999, 2000 Creative Labs, Inc.
 *
 **********************************************************************
 *
 *     Date                 Author          Summary of changes
 *     ----                 ------          ------------------
 *     October 20, 1999     Bertrand Lee    base code release
 *
 **********************************************************************
 *
 *     This program is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU General Public License as
 *     published by the Free Software Foundation; either version 2 of
 *     the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with this program; if not, write to the Free
 *     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 *     USA.
 *
 **********************************************************************
 */

#include "hwaccess.h"
#include "cardwo.h"
#include "audio.h"

/* Volume calcs */

static int set_volume_instance(struct emu10k1_waveout *card_waveout, struct wave_out *wave_out, struct voice_param *left)
{
	/* only applicable for playback */
	u32 volL, volR, vol = 0;

	volL = (wave_out->localvol & 0xffff);
	volR = ((wave_out->localvol >> 16) & 0xffff);

	if (wave_out->globalvolFactor) {
		volL = ((u32) (((u16) card_waveout->globalvol & 0xffff) * (u16) volL)) / 0xffff;
		volR = ((u32) (((u16) (card_waveout->globalvol >> 16) & 0xffff) * ((u16) volR))) / 0xffff;
	}

	/* BIG ASSUMPTION HERE THAT DEFAULT WAVE PAN/AUX IS 0xff/0xff */
	/* New volume and pan */

	if (volL == volR) {
		vol = volL;
		left->unSends.tSends.pan_send = 0xff;
		left->unSends.tSends.aux_send = 0xff;
	} else {
		if (volL > volR) {
			vol = volL;
			left->unSends.tSends.pan_send = 0xff;
			left->unSends.tSends.aux_send = (char) ((volR * 255) / vol);
		} else {
			vol = volR;
			left->unSends.tSends.aux_send = 0xff;
			left->unSends.tSends.pan_send = (char) ((volL * 255) / vol);
		}
	}

	left->initial_attn = 0xff & sumVolumeToAttenuation(vol * 2);

	return vol;
}

static void set_volume(struct emu10k1_waveout *card_waveout)
{
	struct wave_out *currinst = card_waveout->list;
	struct voice_cntlset setting[3];

	while (currinst && currinst->emu_voice) {
		struct voice_param *left = &currinst->emu_voice->voice_params;

		/* This flag allows a wave instance to be unaffected by changes to global volume. */

		if (currinst->globalvolFactor) {
			u32 volL, volR, vol;

			/* Calculate individual channel volumes based on obj and instance volume */

			volL = ((u32) (((u16) card_waveout->globalvol & 0xffff) * ((u16) currinst->localvol & 0xffff))) / 0xffff;
			volR = ((u32) (((u16) (card_waveout->globalvol >> 16) & 0xffff) * ((u16) (currinst->localvol >> 16) & 0xffff))) / 0xffff;

			/* BIG ASSUMPTION HERE THAT DEFAULT WAVE PAN/AUX IS 0xff/0xff */
			/* New volume and pan */
			if (volL == volR) {
				vol = volL;
				left->unSends.tSends.pan_send = 0xff;
				left->unSends.tSends.aux_send = 0xff;
			} else {
				if (volL > volR) {
					vol = volL;
					left->unSends.tSends.pan_send = 0xff;
					left->unSends.tSends.aux_send = (char) ((volR * 255) / vol);
				} else {
					vol = volR;
					left->unSends.tSends.aux_send = 0xff;
					left->unSends.tSends.pan_send = (char) ((volL * 255) / vol);
				}
			}

			left->initial_attn = 0xff & sumVolumeToAttenuation(vol * 2);

			if (currinst->wavexferbuf->is_stereo && currinst->emu_voice->linked_voice) {
				struct voice_param *right = &currinst->emu_voice->linked_voice->voice_params;

				right->unSends.tSends.pan_send = 0;	/* Left output of right channel is always zero */
				right->unSends.tSends.aux_send = left->unSends.tSends.aux_send;	/* Update right channel aux */
				left->unSends.tSends.aux_send = 0;	/* Zero out right output of left channel */
				right->initial_attn = 0xff & sumVolumeToAttenuation(vol * 2);	/* Update right channel attenuation */

				setting[0].paramID = PSST_FXSENDAMOUNT_C;
				setting[0].value = 0;
				setting[1].paramID = PTRX_FXSENDAMOUNT_B;
				setting[1].value = right->unSends.tSends.aux_send;
				setting[2].paramID = IFATN_ATTENUATION;
				setting[2].value = right->initial_attn;

				sblive_voice_setcontrol(currinst->emu_voice->linked_voice, setting, 3);
			}

			setting[0].paramID = PSST_FXSENDAMOUNT_C;
			setting[0].value = left->unSends.tSends.pan_send;
			setting[1].paramID = PTRX_FXSENDAMOUNT_B;
			setting[1].value = left->unSends.tSends.aux_send;
			setting[2].paramID = IFATN_ATTENUATION;
			setting[2].value = left->initial_attn;
			sblive_voice_setcontrol(currinst->emu_voice, setting, 3);
		}

		currinst = currinst->next;
	}
}

static void query_format(struct wave_format *wave_fmt)
{
	if (wave_fmt->channels != 1 && wave_fmt->channels != 2)
		wave_fmt->channels = 2;

	if (wave_fmt->samplingrate >= 0x2EE00)
		wave_fmt->samplingrate = 0x2EE00;

	if (wave_fmt->bitsperchannel != 8 && wave_fmt->bitsperchannel != 16)
		wave_fmt->bitsperchannel = 16;

	return;
}

static int alloc_xferbuffer(struct emu10k1_card *card, struct wave_out *wave_out, u32 * size, void ***buffer)
{
	u32 numpages, reqsize, pagecount;
	struct wave_xferbuf *wavexferbuf = wave_out->wavexferbuf;
	unsigned long busaddx;
	int i;

	reqsize = *size;
	numpages = reqsize / PAGE_SIZE;

	/* If size is not a multiple of PAGE_SIZE then we need to round up */
	if (reqsize % PAGE_SIZE)
		numpages += 1;

	DPD(2, "requested pages is: %d\n", numpages);

	wavexferbuf->numpages = numpages;

	/* Only for playback, request for emu address space */
	/* Support non page-aligned buffer, don't need interpolation page */

	if (emu10k1_addxmgr_alloc(numpages * PAGE_SIZE, card->emupagetable, &wave_out->emuaddr) != CTSTATUS_SUCCESS) {
		DPF(2, "no memory, increase MAXPAGES\n");
		return CTSTATUS_NOMEMORY;
	}

	wave_out->pagetable = (void **) kmalloc(sizeof(void *) * numpages, GFP_KERNEL);

	if (!wave_out->pagetable) {
		emu10k1_addxmgr_free(card->emupagetable, wave_out->emuaddr);
		return CTSTATUS_NOMEMORY;
	}

	/* Fill in virtual memory table */
	for (pagecount = 0; pagecount < numpages; pagecount++) {
		if (!(wave_out->pagetable[pagecount] = (void *) __get_free_page(GFP_KERNEL))) {
			while (pagecount--)
				free_page((unsigned long) wave_out->pagetable[pagecount]);

			emu10k1_addxmgr_free(card->emupagetable, wave_out->emuaddr);
			kfree(wave_out->pagetable);
			return CTSTATUS_NOMEMORY;
		}

		DPD(3, "Virtual Addx: %p\n", wave_out->pagetable[pagecount]);

		for (i = 0; i < PAGE_SIZE / EMUPAGESIZE; i++) {

			busaddx = virt_to_bus((u8 *) wave_out->pagetable[pagecount] + i * EMUPAGESIZE);
			DPD(3, "Bus Addx: %lx\n", busaddx);

			((u32 *) card->virtualpagetable->virtaddx)[(wave_out->emuaddr->emustartaddr >> 11) +
								   pagecount * PAGE_SIZE / EMUPAGESIZE + i] = (((u32) busaddx) & 0xfffff000) << 1;
		}
	}

	*buffer = wave_out->pagetable;

	return CTSTATUS_SUCCESS;
}

static int get_xferbuffer(struct emu10k1_card *card, struct wave_out *wave_out, u32 * size)
{
	struct wave_xferbuf *wavexferbuf = wave_out->wavexferbuf;
	void **buffer;

	wavexferbuf->xferpos = 0;
	wavexferbuf->silence_xferpos = 0;
	wavexferbuf->stopposition = 0;
	wavexferbuf->is_stereo = (wave_out->wave_fmt.channels == 2) ? 1 : 0;
	wavexferbuf->is_16bit = (wave_out->wave_fmt.bitsperchannel == 16) ? 1 : 0;
	wavexferbuf->bytespersample = (wavexferbuf->is_stereo + 1) * (wavexferbuf->is_16bit + 1);

	if (alloc_xferbuffer(card, wave_out, size, &buffer) != CTSTATUS_SUCCESS)
		return CTSTATUS_ERROR;

	/* xferbufsize contains actual transfer buffer size */
	wavexferbuf->xferbufsize = *size;
	wavexferbuf->xferbuffer = buffer;

	return CTSTATUS_SUCCESS;
}

static void dealloc_xferbuffer(struct emu10k1_card *card, struct wave_out *wave_out)
{
	int pagecount;

	/* For playback, free emu address space */
	emu10k1_addxmgr_free(card->emupagetable, wave_out->emuaddr);

	for (pagecount = 0; pagecount < wave_out->wavexferbuf->numpages; pagecount++)
		free_page((unsigned long) wave_out->pagetable[pagecount]);

	kfree(wave_out->pagetable);
	wave_out->wavexferbuf->xferbuffer = NULL;

	return;
}

static int get_voice(struct emu10k1_card *card, struct wave_out *wave_out, int device)
{
	struct emu10k1_waveout *card_waveout = card->waveout;
	struct wave_xferbuf *wavexferbuf = wave_out->wavexferbuf;
	struct voice_allocdesc voice_allocdesc;
	struct voice_param *left, *right;
	u32 size;

	/* Allocate voices here, if no voices available, return error.
	 * Init voice_allocdesc first.*/

	voice_allocdesc.card = card;
	voice_allocdesc.ownertype = VOICEMGR_USAGE_PLAYBACK;

	voice_allocdesc.numvoicereqs = 1;
	voice_allocdesc.flags[0] = 0;

	if (device == 1)
		voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_FXRT2;

	if (wave_out->wave_fmt.channels == 1)
		voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_MONO;

	if (wave_out->wave_fmt.bitsperchannel == 16)
		voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_16BIT;

	voice_allocdesc.flags[0] |= VOICEMGR_FLAGS_PLAYBACK;

	if (sblive_voice_alloc(&voice_allocdesc, &wave_out->emu_voice) != CTSTATUS_SUCCESS)
		return CTSTATUS_ERROR;

	DPD(2, "voicenum -> %d\n", wave_out->emu_voice->voicenum);

	/* voice initialization */

	left = &wave_out->emu_voice->voice_params;

	/* Calculate pitch */
	left->initial_pitch = (u16) (srToPitch(wave_out->wave_fmt.samplingrate) >> 8);

	DPD(2, "Initial pitch --> %x\n", left->initial_pitch);

	/* Easy way out.. gotta calculate value */
	left->pitch_target = 0;
	left->volume_target = 0;
	left->FC_target = 0;

	left->byampl_env_sustain = 0x7f;
	left->byampl_env_decay = 0x7f;

	if (wave_out->globalreverbFactor) {
		u8 t = (card_waveout->globalreverb & 0xff) + (wave_out->localreverb & 0xff);

		left->unSends.tSends.reverb_send = (t > 255) ? 255 : t;
	} else {
		left->unSends.tSends.reverb_send = 0;
	}

	if (wave_out->globalchorusFactor) {
		u8 t = (card_waveout->globalchorus & 0xff) + (wave_out->localchorus & 0xff);

		left->unSends.tSends.chorus_send = (t > 255) ? 255 : t;
	} else {
		left->unSends.tSends.chorus_send = 0;
	}

	set_volume_instance(card_waveout, wave_out, left);

	left->pan_target = left->unSends.tSends.pan_send;
	left->aux_target = left->unSends.tSends.aux_send;

	size = wavexferbuf->xferbufsize / wavexferbuf->bytespersample;
	left->start = 2 * wave_out->emuaddr->emustartaddr / wavexferbuf->bytespersample;
	left->end = left->start + size;
	left->startloop = left->start;
	left->endloop = left->end;

	if (wavexferbuf->is_stereo && wave_out->emu_voice->linked_voice) {
		right = &wave_out->emu_voice->linked_voice->voice_params;

		right->initial_pitch = left->initial_pitch;

		/* Easy way out.. gotta calculate value */
		right->pitch_target = 0;
		right->volume_target = 0;
		right->FC_target = 0;

		right->byampl_env_sustain = 0x7f;
		right->byampl_env_decay = 0x7f;

		right->unSends.tSends.chorus_send = left->unSends.tSends.chorus_send;
		right->unSends.tSends.reverb_send = left->unSends.tSends.reverb_send;

		/* Left output of right channel is always zero */
		right->unSends.tSends.pan_send = 0;

		/* Update right channel aux */
		right->pan_target = 0;
		right->unSends.tSends.aux_send = left->unSends.tSends.aux_send;
		right->aux_target = right->unSends.tSends.aux_send;

		/* Zero out right output of left channel */
		left->unSends.tSends.aux_send = 0;
		left->aux_target = 0;

		/* Update right channel attenuation */
		right->initial_attn = left->initial_attn;

		right->start = left->start;
		right->end = left->end;
		right->startloop = left->startloop;
		right->endloop = left->endloop;

	}

	DPD(2, "voice: start=%x, end=%x, startloop=%x, endloop=%x\n", left->start, left->end, left->startloop, left->endloop);

	return CTSTATUS_SUCCESS;
}

int emu10k1_waveout_open(struct emu10k1_wavedevice *wave_dev)
{
	struct emu10k1_card *card = wave_dev->card;
	struct emu10k1_waveout *card_waveout = card->waveout;
	struct woinst *woinst = wave_dev->woinst;
	struct wave_out *wave_out;
	u32 bytespersec, delay;
	u32 buffsize;

	/* If number of wave instances is greater than the maximum instances allowed, fail the open. */
	if (card_waveout->numplaybackinst + 1 > CARDWAVE_MAXPLAYINST)
		return CTSTATUS_ERROR;

	if (!(wave_out = (struct wave_out *) kmalloc(sizeof(struct wave_out), GFP_KERNEL))) {
		DPF(2, "playback wave instance allocation failed\n");
		return CTSTATUS_NOMEMORY;
	}

	/* Init channel object */
	wave_out->next = NULL;
	wave_out->status = 0;
	wave_out->state = CARDWAVE_STATE_STOPPED;
	wave_out->wave_fmt = woinst->wave_fmt;
	wave_out->emu_voice = NULL;
	wave_out->emuaddr = NULL;
	wave_out->wavexferbuf = NULL;
	wave_out->pagetable = NULL;
	wave_out->timer = NULL;

	/* Assign default local volume */
	/* FIXME: Should we be maxing the initial values like this? */
	wave_out->localvol = 0xffffffff;
	wave_out->localreverb = 0xffffffff;
	wave_out->localchorus = 0xffffffff;
	wave_out->globalvolFactor = 0xffff;
	wave_out->globalreverbFactor = 0xffff;
	wave_out->globalchorusFactor = 0xffff;

	wave_out->setpos = 0;
	wave_out->position = 0;

	wave_out->playflags = CARDWAVE_PLAY_LOOPING;
	wave_out->fill_silence = 0;

	if (!(wave_out->wavexferbuf = (struct wave_xferbuf *) kmalloc(sizeof(struct wave_xferbuf), GFP_KERNEL))) {
		DPF(2, "playback buffer allocation failed\n");
		kfree(wave_out);
		return CTSTATUS_NOMEMORY;
	}

	buffsize = woinst->fragment_size * woinst->numfrags;

	if (get_xferbuffer(card, wave_out, &buffsize) != CTSTATUS_SUCCESS) {
		DPF(2, "get_xferbuffer() failed\n");
		kfree(wave_out->wavexferbuf);
		kfree(wave_out);
		return CTSTATUS_ERROR;
	}

	woinst->fragment_size = buffsize / woinst->numfrags;
	wave_out->callbacksize = woinst->fragment_size;

	if (get_voice(card, wave_out, woinst->device) != CTSTATUS_SUCCESS) {
		dealloc_xferbuffer(card, wave_out);
		kfree(wave_out->wavexferbuf);
		kfree(wave_out);
		return CTSTATUS_ERROR;
	}

	bytespersec = wave_out->wave_fmt.channels * (wave_out->wave_fmt.bitsperchannel >> 3) * (wave_out->wave_fmt.samplingrate);
	delay = (48000 * wave_out->callbacksize) / bytespersec;

	if (emu10k1_timer_install(card, &wave_out->timer, emu10k1_waveout_bh, (unsigned long) wave_dev, delay / 2) != CTSTATUS_SUCCESS) {
		DPF(2, "emu10k1_timer_install() failed\n");
		sblive_voice_free(wave_out->emu_voice);
		dealloc_xferbuffer(card, wave_out);
		kfree(wave_out->wavexferbuf);
		kfree(wave_out);
		return CTSTATUS_ERROR;
	}

	/* Attach this wave instance to the channel object list */
	card_waveout->numplaybackinst++;

	emu10k1_list_attach((struct sblive_list **) &card_waveout->list, (struct sblive_list *) wave_out);

	woinst->wave_out = wave_out;

	return CTSTATUS_SUCCESS;
}

void emu10k1_waveout_close(struct emu10k1_wavedevice *wave_dev)
{
	struct emu10k1_card *card = wave_dev->card;
	struct emu10k1_waveout *card_waveout = card->waveout;
	struct wave_out *wave_out = wave_dev->woinst->wave_out;

	if (wave_out->state != CARDWAVE_STATE_STOPPED)
		emu10k1_waveout_stop(wave_dev);

	emu10k1_list_remove((struct sblive_list **) &card_waveout->list, (struct sblive_list *) wave_out);

	card_waveout->numplaybackinst--;

	emu10k1_timer_uninstall(card, wave_out->timer);

	dealloc_xferbuffer(card, wave_out);
	kfree(wave_out->wavexferbuf);

	if (sblive_voice_free(wave_out->emu_voice) != CTSTATUS_SUCCESS)
		DPF(2, "Failed to free voice\n");

	kfree(wave_out);
	wave_dev->woinst->wave_out = NULL;

	return;
}

int emu10k1_waveout_start(struct emu10k1_wavedevice *wave_dev)
{
	struct emu10k1_card *card = wave_dev->card;
	struct wave_out *wave_out = wave_dev->woinst->wave_out;
	u32 start, startPosition;

	/* If already started, return success */
	if (wave_out->state == CARDWAVE_STATE_STARTED)
		return CTSTATUS_SUCCESS;

	if (wave_out->state == CARDWAVE_STATE_STOPPED && wave_out->setpos)
		startPosition = wave_out->position / (wave_out->wavexferbuf->bytespersample);
	else
		startPosition = wave_out->wavexferbuf->stopposition;

	start = wave_out->emu_voice->voice_params.start;
	wave_out->emu_voice->voice_params.start += startPosition;

	DPD(2, "CA is %x\n", wave_out->emu_voice->voice_params.start);

	if (sblive_voice_playbacksetup(wave_out->emu_voice) != CTSTATUS_SUCCESS)
		return CTSTATUS_ERROR;

	wave_out->emu_voice->voice_params.start = start;

	/* Actual start */
	sblive_voice_start(wave_out->emu_voice);

	wave_out->state = CARDWAVE_STATE_STARTED;
	wave_out->setpos = 0;

	emu10k1_timer_enable(card, wave_out->timer);

	return CTSTATUS_SUCCESS;
}

int emu10k1_waveout_setformat(struct emu10k1_wavedevice *wave_dev)
{
	struct emu10k1_card *card = wave_dev->card;
	struct emu10k1_waveout *card_waveout = card->waveout;
	struct woinst *woinst = wave_dev->woinst;
	struct wave_out *wave_out = woinst->wave_out;
	u32 bytespersec, delay;

	query_format(&woinst->wave_fmt);

	if (!wave_out)
		return CTSTATUS_SUCCESS;

	if (wave_out->state == CARDWAVE_STATE_STARTED) {
		woinst->wave_fmt = wave_out->wave_fmt;
		return CTSTATUS_SUCCESS;
	}

	if ((wave_out->wave_fmt.samplingrate != woinst->wave_fmt.samplingrate)
	    || (wave_out->wave_fmt.bitsperchannel != woinst->wave_fmt.bitsperchannel) || (wave_out->wave_fmt.channels != woinst->wave_fmt.channels)) {
		struct wave_xferbuf *wavexferbuf = wave_out->wavexferbuf;

		emu10k1_timer_uninstall(card, wave_out->timer);

		if (sblive_voice_free(wave_out->emu_voice) != CTSTATUS_SUCCESS)
			DPF(2, "Failed to free voice\n");

		wave_out->wave_fmt = woinst->wave_fmt;

		wavexferbuf->xferpos = 0;
		wavexferbuf->silence_xferpos = 0;
		wavexferbuf->stopposition = 0;
		wavexferbuf->is_stereo = (wave_out->wave_fmt.channels == 2) ? 1 : 0;
		wavexferbuf->is_16bit = (wave_out->wave_fmt.bitsperchannel == 16) ? 1 : 0;
		wavexferbuf->bytespersample = (wavexferbuf->is_stereo + 1) * (wavexferbuf->is_16bit + 1);

		if (get_voice(card, wave_out, woinst->device) != CTSTATUS_SUCCESS) {
			emu10k1_list_remove((struct sblive_list **) &card_waveout->list, (struct sblive_list *) wave_out);
			card_waveout->numplaybackinst--;
			dealloc_xferbuffer(card, wave_out);
			kfree(wavexferbuf);
			kfree(wave_out);
			woinst->wave_out = NULL;
			return CTSTATUS_ERROR;
		}

		bytespersec = wave_out->wave_fmt.channels * (wave_out->wave_fmt.bitsperchannel >> 3) * (wave_out->wave_fmt.samplingrate);
		delay = (48000 * wave_out->callbacksize) / bytespersec;

		if (emu10k1_timer_install(card, &wave_out->timer, emu10k1_waveout_bh, (unsigned long) wave_dev, delay / 2) != CTSTATUS_SUCCESS) {
			DPF(2, "emu10k1_timer_install() failed\n");
			emu10k1_list_remove((struct sblive_list **) &card_waveout->list, (struct sblive_list *) wave_out);
			card_waveout->numplaybackinst--;
			sblive_voice_free(wave_out->emu_voice);
			dealloc_xferbuffer(card, wave_out);
			kfree(wavexferbuf);
			kfree(wave_out);
			woinst->wave_out = NULL;
			return CTSTATUS_ERROR;
		}
	}

	return CTSTATUS_SUCCESS;
}

void emu10k1_waveout_stop(struct emu10k1_wavedevice *wave_dev)
{
	struct emu10k1_card *card = wave_dev->card;
	struct wave_out *wave_out = wave_dev->woinst->wave_out;
	struct wave_xferbuf *wavexferbuf = wave_out->wavexferbuf;
	u32 samples = 32;
	u32 position;

	if (wave_out->state == CARDWAVE_STATE_STOPPED)
		return;

	emu10k1_timer_disable(card, wave_out->timer);

	/* Stop actual voice */
	sblive_voice_stop(wave_out->emu_voice);

	/* Save the stop position */
	sblive_voice_getcontrol(wave_out->emu_voice, CCCA_CURRADDR, &wavexferbuf->stopposition);

	wavexferbuf->stopposition -= wave_out->emu_voice->voice_params.start;

	/* Refer to voicemgr.c, CA is not started at zero.  We need to take this into account. */
	position = wavexferbuf->stopposition * wavexferbuf->bytespersample;

	if (!wavexferbuf->is_16bit)
		samples <<= 1;

	if (wavexferbuf->is_stereo)
		samples <<= 1;

	samples -= 4;

	if (position >= samples * (wavexferbuf->is_16bit + 1))
		position -= samples * (wavexferbuf->is_16bit + 1);
	else
		position += wavexferbuf->xferbufsize - samples * (wavexferbuf->is_16bit + 1);

	wavexferbuf->stopposition = position / wavexferbuf->bytespersample;

	DPD(2, "emu10k1_waveoutstop, position is %x\n", wavexferbuf->stopposition);

	wave_out->state = CARDWAVE_STATE_STOPPED;
	wave_out->setpos = 0;
	wave_out->position = 0;

	return;
}

void emu10k1_waveout_getxfersize(struct wave_out *wave_out, u32 * size, u32 * pending, u32 * curpos)
{
	struct wave_xferbuf *wavexferbuf = wave_out->wavexferbuf;

	/* Get position of current address, this is in no. of bytes in play buffer */
	emu10k1_waveout_getcontrol(NULL, wave_out, WAVECURPOS, curpos);

	if ((*curpos > wavexferbuf->silence_xferpos) || ((*curpos == wavexferbuf->silence_xferpos) && (wave_out->state == CARDWAVE_STATE_STARTED))
	    || ((*curpos == wavexferbuf->silence_xferpos) && (wavexferbuf->silence_xferpos != 0) && (wave_out->state == CARDWAVE_STATE_STOPPED))) {
		*size = *curpos - wavexferbuf->silence_xferpos;
		*pending = wavexferbuf->xferbufsize - *size;
	} else {
		*pending = wavexferbuf->silence_xferpos - *curpos;
		*size = wavexferbuf->xferbufsize - *pending;
	}

	if (wavexferbuf->silence_xferpos != wavexferbuf->xferpos) {
		if (*pending < wave_out->callbacksize) {
			wave_out->fill_silence = 2;
			*pending = 0;
			*size = wavexferbuf->xferbufsize;
			wavexferbuf->xferpos = *curpos;
		} else {
			if (wave_out->fill_silence == 2) {
				*pending = 0;
				*size = wavexferbuf->xferbufsize;
				wavexferbuf->xferpos = *curpos;
			} else {
				*pending -= wave_out->callbacksize;
				*size += wave_out->callbacksize;
			}
		}
	} else {
		if (*pending < wave_out->callbacksize)
			wave_out->fill_silence = 1;
		else
			wave_out->fill_silence = 0;
	}

	return;
}

static void copy_block(u32 dst, u8 * src, u32 len, void **pt)
{
	int i, j, k;

	i = dst / PAGE_SIZE;
	j = dst % PAGE_SIZE;
	k = (len > PAGE_SIZE - j) ? PAGE_SIZE - j : len;
	copy_from_user(pt[i] + j, src, k);
	len -= k;
	while (len >= PAGE_SIZE) {
		copy_from_user(pt[++i], src + k, PAGE_SIZE);
		k += PAGE_SIZE;
		len -= PAGE_SIZE;
	}
	copy_from_user(pt[++i], src + k, len);

	return;
}

static void fill_block(u32 dst, u8 val, u32 len, void **pt)
{
	int i, j, k;

	i = dst / PAGE_SIZE;
	j = dst % PAGE_SIZE;
	k = (len > PAGE_SIZE - j) ? PAGE_SIZE - j : len;
	memset(pt[i] + j, val, k);
	len -= k;
	while (len >= PAGE_SIZE) {
		memset(pt[++i], val, PAGE_SIZE);
		len -= PAGE_SIZE;
	}
	memset(pt[++i], val, len);

	return;
}

void emu10k1_waveout_xferdata(struct woinst *woinst, u8 * data, u32 * size)
{
	struct wave_out *wave_out = woinst->wave_out;
	struct wave_xferbuf *wavexferbuf = wave_out->wavexferbuf;
	u32 sizetocopy, sizetocopy_now, start;
	unsigned long flags;

	sizetocopy = min(wavexferbuf->xferbufsize, *size);
	*size = sizetocopy;

	if (!sizetocopy)
		return;

	spin_lock_irqsave(&woinst->lock, flags);

	sizetocopy_now = wavexferbuf->xferbufsize - wavexferbuf->xferpos;

	start = wavexferbuf->xferpos;

	if (sizetocopy > sizetocopy_now) {
		sizetocopy -= sizetocopy_now;
		wavexferbuf->xferpos = sizetocopy;
		wavexferbuf->silence_xferpos = wavexferbuf->xferpos;
		spin_unlock_irqrestore(&woinst->lock, flags);

		copy_block(start, data, sizetocopy_now, wavexferbuf->xferbuffer);
		copy_block(0, data + sizetocopy_now, sizetocopy, wavexferbuf->xferbuffer);
	} else {
		if (sizetocopy == sizetocopy_now)
			wavexferbuf->xferpos = 0;
		else
			wavexferbuf->xferpos += sizetocopy;

		wavexferbuf->silence_xferpos = wavexferbuf->xferpos;
		spin_unlock_irqrestore(&woinst->lock, flags);

		copy_block(start, data, sizetocopy, wavexferbuf->xferbuffer);
	}

	return;
}

void emu10k1_waveout_fillsilence(struct woinst *woinst)
{
	struct wave_out *wave_out = woinst->wave_out;
	struct wave_xferbuf *wavexferbuf = wave_out->wavexferbuf;
	u16 filldata;
	u32 sizetocopy, sizetocopy_now, start;
	unsigned long flags;

	sizetocopy = wave_out->callbacksize;

	if (wave_out->wave_fmt.bitsperchannel == 8)
		filldata = 0x8080;
	else
		filldata = 0x0000;

	spin_lock_irqsave(&woinst->lock, flags);

	sizetocopy_now = wavexferbuf->xferbufsize - wavexferbuf->silence_xferpos;
	start = wavexferbuf->silence_xferpos;

	if (sizetocopy > sizetocopy_now) {
		sizetocopy -= sizetocopy_now;
		wavexferbuf->silence_xferpos = sizetocopy;
		spin_unlock_irqrestore(&woinst->lock, flags);
		fill_block(start, filldata, sizetocopy_now, wavexferbuf->xferbuffer);
		fill_block(0, filldata, sizetocopy, wavexferbuf->xferbuffer);
	} else {
		if (sizetocopy == sizetocopy_now)
			wavexferbuf->silence_xferpos = 0;
		else
			wavexferbuf->silence_xferpos += sizetocopy;

		spin_unlock_irqrestore(&woinst->lock, flags);

		fill_block(start, filldata, sizetocopy, wavexferbuf->xferbuffer);
	}

	return;
}

/* set the specified control value of the wave device. */

int emu10k1_waveout_setcontrol(struct emu10k1_card *card, struct wave_out *wave_out, u32 ctrl_id, u32 * control_value)
{
	struct emu10k1_waveout *card_waveout = card->waveout;
	struct wave_out *currinst = NULL;
	u32 value;
	struct voice_cntlset setting[6];

	if (card_waveout == NULL)
		return CTSTATUS_ERROR;

	if (control_value)
		value = *control_value;
	else
		value = 0;

	switch (ctrl_id) {
	case WAVEOBJVOLUME:
		return CTSTATUS_SUCCESS;

	case WAVEOBJMUTE:
		/* This is a general mute affecting all wave instances */
		card_waveout->mute = value;
		currinst = card_waveout->list;

		if (value) {
			/* Mute */
			card_waveout->globalvol = 0;
		} else {
			/* Unmute */
			card_waveout->globalvol = (0xffff & card_waveout->left) | ((0xffff & card_waveout->right) << 16);

			set_volume(card_waveout);
		}

		return CTSTATUS_SUCCESS;

	case WAVEOBJREVERB:
		/* This is a general set reverb affecting all wave instances.
		 * Apps which can set the reverb of individual wave instances
		 * should use WAVEINSTANCEREVERB instead. */

		card_waveout->globalreverb = value;
		currinst = card_waveout->list;

		while (currinst && currinst->emu_voice) {
			struct voice_param *left = &currinst->emu_voice->voice_params;

			/* This flag allows a wave instance to be unaffected by changes to global volume. */
			if (currinst->globalreverbFactor) {
				u8 t = (card_waveout->globalreverb & 0xff) + (currinst->localreverb & 0xff);

				left->unSends.tSends.reverb_send = (t > 255) ? 255 : t;

				if (currinst->wavexferbuf->is_stereo && currinst->emu_voice->linked_voice) {
					struct voice_param *right = &currinst->emu_voice->linked_voice->voice_params;

					right->unSends.tSends.reverb_send = left->unSends.tSends.reverb_send;

					setting[0].paramID = PTRX_FXSENDAMOUNT_A;
					setting[0].value = right->unSends.tSends.reverb_send;

					sblive_voice_setcontrol(currinst->emu_voice->linked_voice, setting, 1);
				}

				setting[0].paramID = PTRX_FXSENDAMOUNT_A;
				setting[0].value = left->unSends.tSends.reverb_send;

				sblive_voice_setcontrol(currinst->emu_voice, setting, 1);
			}
			currinst = currinst->next;
		}
		return CTSTATUS_SUCCESS;

	case WAVEOBJCHORUS:
		/* This is a general set chorus affecting all wave instances.
		 * Apps which can set the chorus of individual wave instances
		 * should use WAVEINSTANCECHORUS instead. */

		card_waveout->globalchorus = value;
		currinst = card_waveout->list;

		while (currinst && currinst->emu_voice) {
			struct voice_param *left = &currinst->emu_voice->voice_params;

			if (currinst->globalchorusFactor) {
				u8 t = (card_waveout->globalchorus & 0xff) + (currinst->localchorus & 0xff);

				left->unSends.tSends.chorus_send = (t > 255) ? 255 : t;

				if (currinst->wavexferbuf->is_stereo && currinst->emu_voice->linked_voice) {
					struct voice_param *right = &currinst->emu_voice->linked_voice->voice_params;

					right->unSends.tSends.chorus_send = left->unSends.tSends.chorus_send;

					setting[0].paramID = DSL_FXSENDAMOUNT_D;
					setting[0].value = right->unSends.tSends.chorus_send;

					sblive_voice_setcontrol(currinst->emu_voice->linked_voice, setting, 1);
				}
				setting[0].paramID = DSL_FXSENDAMOUNT_D;
				setting[0].value = left->unSends.tSends.chorus_send;

				sblive_voice_setcontrol(currinst->emu_voice, setting, 1);
			}
			currinst = currinst->next;
		}
		return CTSTATUS_SUCCESS;

	default:
		break;
	}

	if (!wave_out)
		return CTSTATUS_ERROR;

	switch (ctrl_id) {
	case WAVECURPOS:
		if (wave_out->emu_voice == NULL)
			return CTSTATUS_ERROR;

		if (wave_out->state == CARDWAVE_STATE_STOPPED) {
			wave_out->setpos = 1;
			wave_out->position = value;
		} else {
			struct voice_param *left = &wave_out->emu_voice->voice_params;
			struct wave_xferbuf *wavexferbuf = wave_out->wavexferbuf;

			if (wavexferbuf->is_16bit)
				value <<= 1;
			if (wavexferbuf->is_stereo)
				value >>= 1;
			/* get no. of samples per channel */
			value >>= 1;

			/* if the position is already beyond the loop, return error. */
			/* we could check this value before lock can't we ? */
			if (value >= (left->end - left->start))
				return CTSTATUS_ERROR;

			/* Set current position */
			setting[0].paramID = CCCA_CURRADDR;
			setting[0].value = value + left->start;
			sblive_voice_setcontrol(wave_out->emu_voice, setting, 1);

			if (wavexferbuf->is_stereo && wave_out->emu_voice->linked_voice) {
				setting[0].paramID = CCCA_CURRADDR;
				setting[0].value = value + wave_out->emu_voice->linked_voice->voice_params.start;

				sblive_voice_setcontrol(wave_out->emu_voice->linked_voice, setting, 1);
			}
		}
		break;

	case WAVESTOPPOSITION:
		wave_out->wavexferbuf->stopposition = value;
		break;

	case WAVEWRITEPOINTER:
		/* For wave output device, set write pointer for a particular wave instance.
		 * For wave input device, set the offset of write pointer. */
		wave_out->wavexferbuf->xferpos = value;
		break;

	case WAVEINSTANCEVOLUME:
		/* This is a specific set volume affecting only 1 wave instance and can be only called by the owner.
		 * Apps which can set the volume of individual wave instances should use this message. */

		/* update wave instance volume */
		wave_out->localvol = value;

		if (wave_out->emu_voice != NULL) {
			u32 vol;
			struct voice_param *left = &wave_out->emu_voice->voice_params;

			vol = set_volume_instance(card_waveout, wave_out, left);

			if (wave_out->wavexferbuf->is_stereo && wave_out->emu_voice->linked_voice) {
				struct voice_param *right = &wave_out->emu_voice->linked_voice->voice_params;

				right->unSends.tSends.pan_send = 0;	/* Left output of right channel is always zero */
				right->unSends.tSends.aux_send = left->unSends.tSends.aux_send;	/* Update right channel aux */
				left->unSends.tSends.aux_send = 0;	/* Zero out right output of left channel */
				right->initial_attn = 0xff & sumVolumeToAttenuation(vol * 2);	/* Update right channel attenuation */

				setting[0].paramID = IFATN_ATTENUATION;
				setting[0].value = right->initial_attn;

				sblive_voice_setcontrol(wave_out->emu_voice->linked_voice, setting, 1);
			}

			setting[0].paramID = IFATN_ATTENUATION;
			setting[0].value = left->initial_attn;

			sblive_voice_setcontrol(wave_out->emu_voice, setting, 1);
		}
		break;

	case WAVESETFREQUENCY:
		if (wave_out->emu_voice != NULL) {
			struct voice_param *left = &wave_out->emu_voice->voice_params;

			left->initial_pitch = (u16) (srToPitch(value) >> 8);	/* Calculate pitch */

			setting[0].paramID = IP;
			setting[0].value = left->initial_pitch;
			sblive_voice_setcontrol(wave_out->emu_voice, setting, 1);

			if (wave_out->wavexferbuf->is_stereo && wave_out->emu_voice->linked_voice) {
				struct voice_param *right = &wave_out->emu_voice->linked_voice->voice_params;

				right->initial_pitch = left->initial_pitch;
				setting[0].paramID = IP;
				setting[0].value = right->initial_pitch;

				sblive_voice_setcontrol(wave_out->emu_voice->linked_voice, setting, 1);
			}
		}
		wave_out->wave_fmt.samplingrate = value;
		break;

	case WAVESTARTLOOP:
		if (wave_out->emu_voice != NULL) {
			struct voice_param *left = &wave_out->emu_voice->voice_params;

			setting[0].paramID = PSST_LOOPSTARTADDR;
			setting[0].value = left->startloop;

			sblive_voice_setcontrol(wave_out->emu_voice, setting, 1);

			if (wave_out->wavexferbuf->is_stereo && wave_out->emu_voice->linked_voice) {
				struct voice_param *right = &wave_out->emu_voice->linked_voice->voice_params;

				setting[0].paramID = PSST_LOOPSTARTADDR;
				setting[0].value = right->startloop;

				sblive_voice_setcontrol(wave_out->emu_voice->linked_voice, setting, 1);
			}
		}
		break;

	case WAVEVOLFACTOR:
		wave_out->globalvolFactor = value;
		break;

	case WAVEREVERBFACTOR:
		wave_out->globalreverbFactor = value;
		break;

	case WAVECHORUSFACTOR:
		wave_out->globalchorusFactor = value;
		break;

	case WAVESETSTARTFLAG:
		if (wave_out->playflags != value && (wave_out->state == CARDWAVE_STATE_STARTED)) {
			struct emu10k1_card *card = wave_out->emu_voice->card;

			if (value & CARDWAVE_PLAY_LOOPING) {
				sblive_clearstoponloop(card, wave_out->emu_voice->voicenum);
				if (wave_out->emu_voice->linked_voice)
					sblive_clearstoponloop(card, wave_out->emu_voice->linked_voice->voicenum);
			} else {
				sblive_setstoponloop(card, wave_out->emu_voice->voicenum);
				if (wave_out->emu_voice->linked_voice)
					sblive_setstoponloop(card, wave_out->emu_voice->linked_voice->voicenum);
			}
		}
		wave_out->playflags = value;
		break;

	case WAVESETSTOPONLOOP:
		if (wave_out->state == CARDWAVE_STATE_STARTED) {
			struct emu10k1_card *card = wave_out->emu_voice->card;
			u32 CA, offset, endloop;
			struct voice_param *left = &wave_out->emu_voice->voice_params;
			struct voice_cntlset setting[3];

			CA = sblive_readptr(card, CCCA_CURRADDR, wave_out->emu_voice->voicenum) - left->startloop;
			offset = (CA / wave_out->callbacksize) * wave_out->callbacksize;

			if (CA % wave_out->callbacksize)
				offset += wave_out->callbacksize;

			if (CA > (offset - 32))
				offset += 32;

			offset = CA;
			endloop = offset + left->startloop;
			setting[0].paramID = DSL_LOOPENDADDR;
			setting[0].value = endloop;

			sblive_voice_setcontrol(wave_out->emu_voice, setting, 1);

			if (wave_out->wavexferbuf->is_stereo && wave_out->emu_voice->linked_voice) {
				struct voice_param *right = &wave_out->emu_voice->linked_voice->voice_params;

				endloop = offset + right->startloop;
				setting[0].paramID = DSL_LOOPENDADDR;
				setting[0].value = endloop;

				sblive_voice_setcontrol(wave_out->emu_voice, setting, 1);
			}

			sblive_setstoponloop(card, wave_out->emu_voice->voicenum);

			if (wave_out->emu_voice->linked_voice)
				sblive_setstoponloop(card, wave_out->emu_voice->linked_voice->voicenum);
		}
		break;

	default:
		return CTSTATUS_NOTSUPPORTED;
	}

	return CTSTATUS_SUCCESS;
}

/* get the specified control value of the wave device. */

int emu10k1_waveout_getcontrol(struct emu10k1_waveout *card_waveout, struct wave_out *wave_out, u32 ctrl_id, u32 * value)
{
	switch (ctrl_id) {
	case WAVEOBJVOLUME:
		*value = card_waveout->globalvol;
		return CTSTATUS_SUCCESS;

	case WAVEOBJREVERB:
		*value = card_waveout->globalreverb;
		return CTSTATUS_SUCCESS;

	case WAVEOBJCHORUS:
		*value = card_waveout->globalchorus;
		return CTSTATUS_SUCCESS;

	case WAVEQUERYACTIVEINST:
		if (card_waveout->list != NULL)
			return CTSTATUS_SUCCESS;
		else
			return CTSTATUS_ERROR;

	default:
		break;
	}

	if (!wave_out)
		return CTSTATUS_ERROR;

	switch (ctrl_id) {
	case WAVECURPOS:
		/* There is no actual start yet */
		if (wave_out->state == CARDWAVE_STATE_STOPPED) {
			if (wave_out->setpos)
				*value = wave_out->position;
			else
				*value = wave_out->wavexferbuf->stopposition * wave_out->wavexferbuf->bytespersample;
		} else {
			sblive_voice_getcontrol(wave_out->emu_voice, CCCA_CURRADDR, value);

			*value -= wave_out->emu_voice->voice_params.start;

			/* Get number of bytes in play buffer per channel.
			 * If 8 bit mode is enabled, this needs to be changed. */
			{
				u32 samples = 64 * (wave_out->wavexferbuf->is_stereo + 1);

				*value *= wave_out->wavexferbuf->bytespersample;

				/* Refer to voicemgr.c, CA is not started at zero.
				 * We need to take this into account. */

				samples -= 4 * (wave_out->wavexferbuf->is_16bit + 1);

				if (*value >= samples)
					*value -= samples;
				else
					*value += wave_out->wavexferbuf->xferbufsize - samples;
			}
		}

		break;

	case WAVEWRITEPOINTER:
		/* Get write pointer for a particular wave instance */
		*value = wave_out->wavexferbuf->xferpos;
		break;

	case WAVEINSTANCEVOLUME:
		*value = wave_out->localvol;
		return CTSTATUS_SUCCESS;

	case WAVEINSTANCEREVERB:
		*value = wave_out->localreverb;
		return CTSTATUS_SUCCESS;

	case WAVEINSTANCECHORUS:
		*value = wave_out->localchorus;
		return CTSTATUS_SUCCESS;

	case WAVESTARTLOOP:
		if (wave_out->emu_voice == NULL)
			return CTSTATUS_ERROR;

		sblive_voice_getcontrol(wave_out->emu_voice, PSST_LOOPSTARTADDR, value);

		*value -= wave_out->emu_voice->voice_params.start;

		/* Get number of bytes in play buffer per channel.
		 * If 8 bit mode is enabled, this needs to be changed. */
		*value *= wave_out->wavexferbuf->bytespersample;
		break;

	case WAVEENDLOOP:
		if (wave_out->emu_voice == NULL)
			return CTSTATUS_ERROR;

		sblive_voice_getcontrol(wave_out->emu_voice, DSL_LOOPENDADDR, value);

		*value -= wave_out->emu_voice->voice_params.start;

		/* Get number of bytes in play buffer per channel.
		 * Iif 8 bit mode is enabled, this needs to be changed. */
		*value *= wave_out->wavexferbuf->bytespersample;
		break;

	default:
		return CTSTATUS_NOTSUPPORTED;
	}

	return CTSTATUS_SUCCESS;
}
