
/*
 **********************************************************************
 *     timer.c
 *     Copyright (C) 1999, 2000 Creative Labs, inc.
 *
 **********************************************************************
 *
 *     This program is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU General Public License as
 *     published by the Free Software Foundation; either version 2 of
 *     the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with this program; if not, write to the Free
 *     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 *     USA.
 *
 **********************************************************************
 */

/* 3/6/2000	Improved support for different timer delays  Rui Sousa */

#include "hwaccess.h"

/* Try to schedule only once per fragment */

void emu10k1_timer_irqhandler(struct emu10k1_card *card)
{
	struct emu_timer *t;

	spin_lock(&card->timer_lock);

	t = card->timer;
	while (t != NULL) {
		if (t->active) {
			t->count++;
			if (t->count == t->count_max) {
				t->count = 0;
				tasklet_hi_schedule(&t->tasklet);
			}
		}
		t = t->next;
	}

	spin_unlock(&card->timer_lock);

	return;
}

int emu10k1_timer_install(struct emu10k1_card *card, struct emu_timer **timer, void (*func) (unsigned long), unsigned long data, u32 delay)
{
	struct emu_timer *t;
	unsigned long flags;

	t = (struct emu_timer *) kmalloc(sizeof(struct emu_timer), GFP_KERNEL);

	if (t == NULL) {
		DPF(2, "timer alloc failed\n");
		return CTSTATUS_NOMEMORY;
	}

	DPD(3, "kmalloc: [%p]\n", t);

	if (delay < 5)
		delay = 5;

	t->delay = delay;
	tasklet_init(&t->tasklet, func, data);
	t->active = 0;

	*timer = t;

	spin_lock_irqsave(&card->timer_lock, flags);

	t->count_max = t->delay / (card->timer_delay < 1024 ? card->timer_delay : 1024);
	t->count = t->count_max - 1;
	t->next = card->timer;
	card->timer = t;

	if (card->timer_delay > delay) {
		if (card->timer_delay == TIMER_STOPPED)
			sblive_irqmgr_enableirq(card, INTE_INTERVALTIMERENB);

		card->timer_delay = delay;
		delay = (delay < 1024 ? delay : 1024);

		WRITE_FN0(card, TIMER_RATE, delay);

		t = card->timer;
		while (t != NULL) {
			t->count_max = t->delay / delay;
			/* don't want to think much, just force scheduling 
			   on the next interrupt */
			t->count = t->count_max - 1;
			t = t->next;
		}

		DPD(2, "timer rate --> %u\n", delay);
	}

	spin_unlock_irqrestore(&card->timer_lock, flags);

	return CTSTATUS_SUCCESS;
}

void emu10k1_timer_uninstall(struct emu10k1_card *card, struct emu_timer *timer)
{
	struct emu_timer *t;
	u32 delay = TIMER_STOPPED;
	unsigned long flags;

	spin_lock_irqsave(&card->timer_lock, flags);

	if (card->timer == timer) {
		card->timer = timer->next;
	}

	if ((t = card->timer) != NULL) {
		delay = t->delay;

		while (t->next != NULL) {
			if (t->next == timer) {
				t->next = timer->next;
				continue;
			}

			if (t->next->delay < delay)
				delay = t->next->delay;

			t = t->next;
		}
	}

	if (card->timer_delay != delay) {
		card->timer_delay = delay;

		if (delay == TIMER_STOPPED)
			sblive_irqmgr_disableirq(card, INTE_INTERVALTIMERENB);
		else {
			delay = (delay < 1024 ? delay : 1024);

			WRITE_FN0(card, TIMER_RATE, delay);

			t = card->timer;
			while (t != NULL) {
				t->count_max = t->delay / delay;
				t->count = t->count_max - 1;
				t = t->next;
			}
		}

		DPD(2, "timer rate --> %u\n", delay);
	}

	spin_unlock_irqrestore(&card->timer_lock, flags);

	tasklet_unlock_wait(&timer->tasklet);
	kfree(timer);
	DPD(3, "kfree: [%p]\n", timer);

	return;
}

void emu10k1_timer_enable(struct emu10k1_card *card, struct emu_timer *timer)
{
	unsigned long flags;

	spin_lock_irqsave(&card->timer_lock, flags);
	timer->active = 1;
	spin_unlock_irqrestore(&card->timer_lock, flags);

	return;
}

void emu10k1_timer_disable(struct emu10k1_card *card, struct emu_timer *timer)
{
	unsigned long flags;

	spin_lock_irqsave(&card->timer_lock, flags);
	timer->active = 0;
	spin_unlock_irqrestore(&card->timer_lock, flags);

	return;
}
