
/*
 **********************************************************************
 *     sblive_voice.c - Voice manager for emu10k1 driver
 *     Copyright 1999, 2000 Creative Labs, Inc.
 *
 **********************************************************************
 *
 *     Date                 Author          Summary of changes
 *     ----                 ------          ------------------
 *     October 20, 1999     Bertrand Lee    base code release
 *
 **********************************************************************
 *
 *     This program is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU General Public License as
 *     published by the Free Software Foundation; either version 2 of
 *     the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with this program; if not, write to the Free
 *     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 *     USA.
 *
 **********************************************************************
 */

#include "hwaccess.h"

static struct emu_voice *voice_alloc(struct voice_mgr *voice, u32 flags)
{
	unsigned int voicenum;
	struct emu10k1_card *card = voice->card;
	struct emu_voice *foundvoice = NULL;
	struct emu_voice *slavevoice = NULL;
	struct emu_voice *testvoice, *testslave;
	u32 curvol, lowestvol, mastervol, slavevol, CA, dcysusv;

	DPF(2, "Entered voice_alloc\n");

	if (flags & VOICEMGR_FLAGS_MONO) {
		DPF(2, "voice_alloc MONO PLAYBACK voice\n");

		/* Check for free voices first */
		if (voice->free_voices) {
			voice->free_voices->voicestate = VOICEMGR_STATE_TRANSITION;
			foundvoice = voice->free_voices;
			foundvoice->flags = 0;
			DPD(2, "  Voice found %d\n", foundvoice->voicenum);
			return foundvoice;
		}

		/* We couldn't find any free voices so we need to free up
		 * some voices. We only rip off MIDI voices at the moment. */
		testvoice = dlFirstNode(&voice->midi_voices);

		lowestvol = 0xffffffff;

		while (testvoice) {
			/* Voice stealing algorithm goes here */
			voicenum = testvoice->voicenum;

			if (!(testvoice->voicestate & VOICEMGR_STATE_TRANSITION)) {
				/* Check if voice is looping silence */
				CA = sblive_readptr(card, CCCA_CURRADDR, voicenum);
				if (CA > testvoice->voice_params.end) {
					testvoice->voicestate |= VOICEMGR_STATE_TRANSITION;
					if (foundvoice)
						foundvoice->voicestate &= ~VOICEMGR_STATE_TRANSITION;
					lowestvol = 0;
					foundvoice = testvoice;
					break;
				}

				curvol = sblive_readptr(card, CVCF_CURRENTVOL, voicenum);
				dcysusv = sblive_readptr(card, DCYSUSV, voicenum);
				dcysusv = (dcysusv & DCYSUSV_PHASE1_MASK) ? 0 : 0x1000;	/* 0x1000 means what? */

				if ((curvol < 0x200) && !dcysusv && (testvoice->voicestate & VOICEMGR_STATE_RELEASE)) {
					/* The volume is less than the threshold and
					   the volume envelope is in the decaying state.
					   Stop the pitch to stop the busmaster activities.
					 */
					sblive_writeptr(card, IP, voicenum, 0);
					sblive_writeptr(card, VTFT, voicenum, VTFT_FILTERTARGET_MASK);
					sblive_writeptr(card, PTRX_PITCHTARGET, voicenum, 0);
				} else {
					if (testvoice->voicestate & VOICEMGR_STATE_SUSTAIN)
						curvol += 0x500 + dcysusv;
					else if (testvoice->voicestate & VOICEMGR_STATE_ATTACK)
						curvol += 0x200 + dcysusv;
				}

				if (curvol < lowestvol) {
					testvoice->voicestate |= VOICEMGR_STATE_TRANSITION;

					if (foundvoice)
						foundvoice->voicestate &= ~VOICEMGR_STATE_TRANSITION;
					foundvoice = testvoice;
					lowestvol = curvol;
				}
			}
			testvoice = dlNextNode(testvoice);
		}

		if (foundvoice) {
			foundvoice->flags = 0;
			foundvoice->master_voice = NULL;
			foundvoice->linked_voice = NULL;
		} else {
			DPF(2, "Voice not found!\n");
		}

		return foundvoice;
	} else {
		DPF(2, "voice_alloc STEREO PLAYBACK voice\n");

		voicenum = 8;
		lowestvol = 0xffffffff;

		do {
			/* Since stereo voices must be side-by-side,
			   we have to check 2 voices at once. */
			if (((voice->voices[voicenum].usage == VOICEMGR_USAGE_FREE)
			     || (voice->voices[voicenum].usage == VOICEMGR_USAGE_MIDI))
			    && ((voice->voices[voicenum + 1].usage == VOICEMGR_USAGE_FREE)
				|| (voice->voices[voicenum + 1].usage == VOICEMGR_USAGE_MIDI))) {
				mastervol = 0xffffffff;
				slavevol = 0xffffffff;
				testvoice = &voice->voices[voicenum];
				testslave = &voice->voices[voicenum + 1];

				if ((!(testvoice->voicestate & VOICEMGR_STATE_TRANSITION))
				    && (!(testslave->voicestate & VOICEMGR_STATE_TRANSITION))) {
					/* Voice stealing algorithm goes here */

					/* Check master voice */
					if (testvoice->usage == VOICEMGR_USAGE_FREE)
						mastervol = 0;
					else {
						/* Check if voice is looping silence */
						CA = sblive_readptr(card, CCCA_CURRADDR, voicenum);

						if (CA > testvoice->voice_params.end)
							mastervol = 0;
						else {
							mastervol = sblive_readptr(card, CVCF_CURRENTVOL, voicenum);
							if (!(testvoice->voicestate & VOICEMGR_STATE_RELEASE))
								mastervol += 0x1300;	/* Why??? */
						}
					}

					/* Check slave voice */
					if (testslave->usage == VOICEMGR_USAGE_FREE)
						slavevol = 0;
					else {
						/* Check if voice is looping silence */
						CA = sblive_readptr(card, CCCA_CURRADDR, voicenum + 1);

						if (CA > testslave->voice_params.end)
							slavevol = 0;
						else {
							slavevol = sblive_readptr(card, CVCF_CURRENTVOL, (voicenum + 1));
							if (!(testslave->voicestate & VOICEMGR_STATE_RELEASE))
								slavevol += 0x1300;	/* Why??? */
						}
					}

					if ((mastervol + slavevol) < lowestvol) {
						testvoice->voicestate |= VOICEMGR_STATE_TRANSITION;
						testslave->voicestate |= VOICEMGR_STATE_TRANSITION;

						if (foundvoice) {
							foundvoice->voicestate &= ~VOICEMGR_STATE_TRANSITION;
							slavevoice->voicestate &= ~VOICEMGR_STATE_TRANSITION;
						}

						foundvoice = testvoice;
						slavevoice = testslave;
						lowestvol = mastervol + slavevol;
					}
				}
			}

			voicenum = (voicenum + 2) % NUM_G;

		} while (voicenum != 8);

		if (foundvoice) {
			DPD(2, "Voice found %x\n", (u32) voicenum);

			/* Link next voice in stereo pair */
			foundvoice->master_voice = NULL;
			foundvoice->linked_voice = slavevoice;
			slavevoice->master_voice = foundvoice;
			slavevoice->linked_voice = NULL;

			/* Mark second voice as stereo slave */
			foundvoice->flags = VOICEMGR_FLAGS_VOICEMASTER;
			slavevoice->flags = VOICEMGR_FLAGS_STEREOSLAVE;

			return foundvoice;
		}
	}

	return foundvoice;
}

int sblive_voice_alloc(struct voice_allocdesc *voiceallocdesc, struct emu_voice **voices)
{
	struct emu_voice **ownerlist = NULL;	/* Linux: added initialization */
	struct emu_voice *allocvoicelist;
	struct emu_voice *lastvoice = NULL;	/* Linux: added initialization */
	struct emu_voice *linkvoice;
	int status = CTSTATUS_SUCCESS;
	struct emu10k1_card *card = voiceallocdesc->card;
	unsigned long flags;
	int i;

	if (!voiceallocdesc->numvoicereqs) {
		/* No voices requested??? */
		DPF(2, "No voices requested!\n");
		return CTSTATUS_INVALIDPARAM;
	}

	/* Determine 'busy' list to attach voices to */
	if (voiceallocdesc->ownertype == VOICEMGR_USAGE_PLAYBACK)
		ownerlist = &card->voice_manager.playback_voices;
	else if (voiceallocdesc->ownertype == VOICEMGR_USAGE_MIDI)
		ownerlist = &card->voice_manager.midi_voices;

	allocvoicelist = NULL;

	/* Satisfy voice requests one by one */
	for (i = 0; i < voiceallocdesc->numvoicereqs; i++) {
		/* Allocate voices */

		linkvoice = voice_alloc(&card->voice_manager, voiceallocdesc->flags[i]);

		if (linkvoice) {
			/* Suitable free voice found */
			if (allocvoicelist == NULL) {
				allocvoicelist = linkvoice;
				linkvoice->master_voice = NULL;
			} else {
				lastvoice->linked_voice = linkvoice;
				linkvoice->master_voice = lastvoice;
			}

			linkvoice->flags = VOICEMGR_FLAGS_VOICEMASTER;

			/* Repeat for all linked voices */
			while (linkvoice) {
				/* Shut off the voice */
				sblive_writeptr(card, IFATN, linkvoice->voicenum, 0xffff);
				sblive_writeptr(card, DCYSUSV, linkvoice->voicenum, ENV_OFF);
				sblive_writeptr(card, VTFT, linkvoice->voicenum, 0xffff);
				sblive_writeptr(card, PTRX, linkvoice->voicenum, 0);

				spin_lock_irqsave(&sblive_spinlock, flags);

				dlDelNode(&card->voice_manager.free_voices, linkvoice);	/* Remove voice from free list */

				linkvoice->card = card;
				linkvoice->ownertype = voiceallocdesc->ownertype;

				/* We need a transition state to make a voice
				   as belonging to a owner but not yet assigned
				   to that owner yet */
				linkvoice->voicestate |= VOICEMGR_STATE_TRANSITION;
				linkvoice->usage = voiceallocdesc->ownertype;
				linkvoice->flags |= voiceallocdesc->flags[i];

				dlAddNode(ownerlist, linkvoice);	/* Attach voice to busy list */

				spin_unlock_irqrestore(&sblive_spinlock, flags);

				lastvoice = linkvoice;	/* Remember last of linked voices */

#ifdef PRIVATE_PCM_VOLUME
				{
					int j;

					for (j = 0; j < MAX_PCM_CHANNELS; j++) {
						if (sblive_pcm_volume[j].files == current->files) {
							if (linkvoice == allocvoicelist) {
								sblive_pcm_volume[j].channel_l = linkvoice->voicenum;
								sblive_pcm_volume[j].channel_r = NUM_G;
								DPD(2, "left voice allocated: %d\n", linkvoice->voicenum);

							} else {
								sblive_pcm_volume[j].channel_r = linkvoice->voicenum;
								DPD(2, "right voice allocated: %d\n", linkvoice->voicenum);
							}
							break;
						}
					}
				}
#endif
				linkvoice = linkvoice->linked_voice;
			}
		} else {
			/* No free suitable voices found - steal from busy voices */
			status = CTSTATUS_INUSE;

			/* TODO : IMPLEMENT VOICE STEALING ALGORITHM */

			DPF(2, "Steal a voice: not yet implemented\n");

			spin_lock_irqsave(&sblive_spinlock, flags);

			while (allocvoicelist) {
				linkvoice = allocvoicelist;
				allocvoicelist = allocvoicelist->linked_voice;
				dlDelNode(ownerlist, linkvoice);

				linkvoice->linked_voice = NULL;
				linkvoice->master_voice = NULL;
				linkvoice->flags = 0;
				linkvoice->voicestate = VOICEMGR_STATE_IDLE;
				linkvoice->usage = VOICEMGR_USAGE_FREE;

				dlAddNode(&card->voice_manager.free_voices, linkvoice);
			}

			spin_unlock_irqrestore(&sblive_spinlock, flags);

			break;
		}

	}

	if (voiceallocdesc->ownertype != VOICEMGR_USAGE_MIDI) {
		linkvoice = allocvoicelist;

		while (linkvoice) {
			linkvoice->voicestate &= ~VOICEMGR_STATE_TRANSITION;
			linkvoice = linkvoice->linked_voice;
		}
	}

	*voices = allocvoicelist;

	return status;
}

/* Frees a previously allocated voice. Multiple voices can
 * be allocated at once but must be freed individually. */

int sblive_voice_free(struct emu_voice *voices)
{
	struct emu_voice *voicetemp;
	struct emu_voice **ownerlist;
	struct emu10k1_card *card;
	unsigned long flags;

	if (voices == NULL)
		return CTSTATUS_INVALIDPARAM;

	card = voices->card;

	if (voices->usage == VOICEMGR_USAGE_PLAYBACK)
		ownerlist = &card->voice_manager.playback_voices;
	else if (voices->usage == VOICEMGR_USAGE_MIDI)
		ownerlist = &card->voice_manager.midi_voices;
	else {
		DPF(2, "Unknown VOICEMGR_USAGE\n");
		return CTSTATUS_ERROR;
	}

	if (!(voices->flags & VOICEMGR_FLAGS_VOICEMASTER)) {
		/* Not first voice in list */
		DPF(2, "Not first voice in list!\n");
		return CTSTATUS_ERROR;
	}

	while (voices) {
		unsigned dcysusv, voicenum;

		voicenum = voices->voicenum;

		sblive_writeptr(card, IFATN, voicenum, IFATN_FILTERCUTOFF_MASK | IFATN_ATTENUATION_MASK);
		sblive_writeptr(card, IP, voicenum, 0);

		dcysusv = sblive_readptr(card, DCYSUSV, voicenum) & (DCYSUSV_PHASE1_MASK | DCYSUSV_SUSTAINLEVEL_MASK | DCYSUSV_DECAYTIME_MASK);
		sblive_writeptr(card, DCYSUSV, voicenum, dcysusv | ENV_OFF);

		sblive_writeptr(card, VTFT, voicenum, VTFT_FILTERTARGET_MASK);
		sblive_writeptr(card, PTRX_PITCHTARGET, voicenum, 0);
		sblive_writeptr(card, CVCF, voicenum, CVCF_CURRENTFILTER_MASK);
		sblive_writeptr(card, CPF, voicenum, 0);

		{
			u32 cra;
			u32 sample;
			int i;

			sample = (voices->flags & VOICEMGR_FLAGS_16BIT) ? 0 : 0x80808080;
			cra = sblive_readptr(card, CCR, voicenum) & CCR_READADDRESS_MASK;
			sblive_writeptr(card, CCR, voicenum, cra);
			cra = (cra >> 18) & 0xf;
			sblive_writeptr(card, CD0 + cra, voicenum, sample);
			cra = (cra + 0x1) & 0xf;
			sblive_writeptr(card, CD0 + cra, voicenum, sample);

			for (i = 0; i < NUM_FXSENDS; i++)
				if (voices->sendhandle[i])
					voices->sendhandle[i] = 0;
		}

		spin_lock_irqsave(&sblive_spinlock, flags);

		/* Reattach to free list */
		voicetemp = voices;
		voices = voices->linked_voice;

		dlDelNode(ownerlist, voicetemp);	/* Detach first */
		voicetemp->linked_voice = NULL;
		voicetemp->master_voice = NULL;
		voicetemp->flags = 0;
		voicetemp->voicestate = VOICEMGR_STATE_IDLE;
		voicetemp->usage = VOICEMGR_USAGE_FREE;
		dlAddNode(&card->voice_manager.free_voices, voicetemp);

		spin_unlock_irqrestore(&sblive_spinlock, flags);

	}

	return CTSTATUS_SUCCESS;
}

/*       Sets up a voices for Wave Playback */

int sblive_voice_playbacksetup(struct emu_voice *voices)
{
	int voicenum;
	struct emu10k1_card *card;
	u32 cra = 0;
	u32 start = 0;
	struct voice_param *voice_params;
	struct emu_voice *currvoice;

	currvoice = voices;
	while (currvoice) {
		/* Or here? (voices can be on different cards) */
		card = currvoice->card;
		voicenum = currvoice->voicenum;
		voice_params = &currvoice->voice_params;

		sblive_writeptr(card, DCYSUSV, voicenum, ENV_OFF);
		sblive_writeptr(card, VTFT, voicenum, VTFT_FILTERTARGET_MASK);
		sblive_writeptr(card, CVCF, voicenum, CVCF_CURRENTFILTER_MASK);

		sblive_writeptr(card, FXRT, voicenum, (currvoice->flags & VOICEMGR_FLAGS_FXRT2) ? 0xd23c0000 : 0xd01c0000);

		/* Stop CA */
		/* Assumption that PT is already 0 so no harm overwriting */
		sblive_writeptr(card, PTRX, voicenum, (voice_params->unSends.tSends.reverb_send << 8) | voice_params->unSends.tSends.aux_send);

		if (currvoice->flags & VOICEMGR_FLAGS_VOICEMASTER) {
			u32 sample;

			if (currvoice->linked_voice != NULL) {
				/* Set stereo bit */
				cra = 64;
				sblive_writeptr(card, CPF, voicenum, CPF_STEREO_MASK);
				sblive_writeptr(card, CPF, voicenum + 1, CPF_STEREO_MASK);
			} else {
				cra = 32;
				sblive_writeptr(card, CPF, voicenum, 0);
			}

			if (currvoice->flags & VOICEMGR_FLAGS_16BIT)
				sample = 0;
			else {
				cra = cra * 2;
				sample = 0x80808080;
			}
			cra -= 4;

			if (currvoice->linked_voice != NULL) {
				/* CCR_READADDRESS_MASK */
				sblive_writeptr(card, CCR, voicenum, 0x3c << 16);
				sblive_writeptr(card, CCR, voicenum + 1, cra << 16);
				sblive_writeptr(card, CDE, voicenum + 1, sample);
				sblive_writeptr(card, CDF, voicenum + 1, sample);
				start = voice_params->start + cra / 2;
			} else {
				sblive_writeptr(card, CCR, voicenum, 0x1c << 16);	/* FIXME: Is 0x1c correct? */
				sblive_writeptr(card, CDE, voicenum, sample);
				sblive_writeptr(card, CDF, voicenum, sample);
				start = voice_params->start + cra;
			}

			if (start > voice_params->endloop) {
				start -= voice_params->endloop;

				if (currvoice->linked_voice != NULL)
					cra = (cra << 25) | 0x1bc0000 | ((cra - start) << 9);
				else
					cra = (cra << 25) | 0x11c0000 | ((cra - start) << 9);

				start += voice_params->startloop;

				if (start >= voice_params->endloop)
					start = voice_params->endloop - 1;
			} else if (currvoice->linked_voice != NULL)
				cra = (cra << 25) | (0x3c << 16);
			else
				cra = (cra << 25) | (0x1c << 16);

			start |= CCCA_INTERPROM_0;
		}

		/* CSL, ST, CA */
		sblive_writeptr(card, DSL, voicenum, voice_params->endloop | ((u32) voice_params->unSends.tSends.chorus_send << 24));

		sblive_writeptr(card, PSST, voicenum, voice_params->startloop | ((u32) voice_params->unSends.tSends.pan_send << 24));

		if (currvoice->flags & VOICEMGR_FLAGS_16BIT)
			sblive_writeptr(card, CCCA, voicenum, start);
		else
			sblive_writeptr(card, CCCA, voicenum, start | CCCA_8BITSELECT);

		/* Clear filter delay memory */
		sblive_writeptr(card, Z1, voicenum, 0);
		sblive_writeptr(card, Z2, voicenum, 0);

		/* Invalidate maps */
		sblive_writeptr(card, MAPA, voicenum, MAP_PTI_MASK | (card->silentpage->busaddx * 2));
		sblive_writeptr(card, MAPB, voicenum, MAP_PTI_MASK | (card->silentpage->busaddx * 2));

		/* Fill cache */
		if (currvoice->flags & VOICEMGR_FLAGS_VOICEMASTER)
			sblive_writeptr(card, CCR, voicenum, cra);

		sblive_writeptr(card, ATKHLDV, voicenum, ATKHLDV_HOLDTIME_MASK | ATKHLDV_ATTACKTIME_MASK);
		sblive_writeptr(card, LFOVAL1, voicenum, 0x8000);
		sblive_writeptr(card, ATKHLDM, voicenum, 0);
		sblive_writeptr(card, DCYSUSM, voicenum, DCYSUSM_DECAYTIME_MASK);
		sblive_writeptr(card, LFOVAL2, voicenum, 0x8000);
		sblive_writeptr(card, IP, voicenum, voice_params->initial_pitch);
		sblive_writeptr(card, PEFE, voicenum, 0x7f);
		sblive_writeptr(card, FMMOD, voicenum, 0);
		sblive_writeptr(card, TREMFRQ, voicenum, 0);
		sblive_writeptr(card, FM2FRQ2, voicenum, 0);
		sblive_writeptr(card, ENVVAL, voicenum, 0xbfff);
		sblive_writeptr(card, ENVVOL, voicenum, 0xbfff);

#ifdef PRIVATE_PCM_VOLUME
		{
			int i;

			for (i = 0; i < MAX_PCM_CHANNELS; i++) {
				if (sblive_pcm_volume[i].channel_l == voicenum) {
					voice_params->initial_attn = (sblive_pcm_volume[i].channel_r < NUM_G) ? sblive_pcm_volume[i].attn_l :
					    // test for mono channel (reverse logic is correct here!)
					    (sblive_pcm_volume[i].attn_r > sblive_pcm_volume[i].attn_l) ?
					    sblive_pcm_volume[i].attn_l : sblive_pcm_volume[i].attn_r;
					DPD(2, "set left volume  %d\n", voice_params->initial_attn);
					break;
				} else if (sblive_pcm_volume[i].channel_r == voicenum) {
					voice_params->initial_attn = sblive_pcm_volume[i].attn_r;
					DPD(2, "set right volume  %d\n", voice_params->initial_attn);
					break;
				}
			}
		}
#endif
		sblive_writeptr(card, IFATN, voicenum, IFATN_FILTERCUTOFF_MASK | voice_params->initial_attn);

		voice_params->FC_target = 0xffff;
		voice_params->pitch_target = (u16) (IP_TO_CP(voice_params->initial_pitch) >> 16);

		currvoice = currvoice->linked_voice;
	}

	return CTSTATUS_SUCCESS;
}

void sblive_voice_start(struct emu_voice *voices)
{
	struct emu10k1_card *card;
	struct emu_voice *tempvoice;

	tempvoice = voices;
	card = tempvoice->card;

	/* Actual start */
	while (tempvoice != NULL) {
		int voicenum = 0;	/* Linux: Added initialization */

		if (tempvoice->flags & VOICEMGR_FLAGS_PLAYBACK) {
			struct voice_param *pvoice_params = &tempvoice->voice_params;

			voicenum = tempvoice->voicenum;

			/* Playback */
			sblive_writeptr(card, PTRX_PITCHTARGET, voicenum, pvoice_params->pitch_target);

			if (!(tempvoice->flags & VOICEMGR_FLAGS_STEREOSLAVE))
				sblive_writeptr(card, CPF_CURRENTPITCH, voicenum, pvoice_params->pitch_target);

			sblive_writeptr(card, VTFT, voicenum, ((u32) pvoice_params->volume_target << 16)
					| pvoice_params->FC_target);
			sblive_writeptr(card, CVCF, voicenum, ((u32) pvoice_params->volume_target << 16)
					| pvoice_params->FC_target);
			sblive_writeptr(card, DCYSUSV, voicenum, (pvoice_params->byampl_env_sustain << 8)
					| ENV_ON | pvoice_params->byampl_env_decay);

			/* Using StopOnLoop for MIDI stops the playback
			   too early, which may cause a DC level to be played
			   until the note is released. */

			if (tempvoice->usage == VOICEMGR_USAGE_MIDI)
				sblive_clearstoponloop(card, voicenum);
			else {
				if (pvoice_params->startloop > pvoice_params->end)
					sblive_setstoponloop(card, voicenum);
				else
					sblive_clearstoponloop(card, voicenum);
			}
		} else
			/* Recording */
			sblive_writeptr(card, DCYSUSV, voicenum, DCYSUSV_SUSTAINLEVEL_MASK | DCYSUSV_DECAYTIME_MASK | ENV_ON);

		tempvoice->voicestate |= VOICEMGR_STATE_ATTACK;
		tempvoice = tempvoice->linked_voice;
	}

	return;
}

void sblive_voice_stop(struct emu_voice *voices)
{
	struct emu10k1_card *card = voices->card;
	struct emu_voice *tempvoice = voices;
	unsigned long flags;

	spin_lock_irqsave(&sblive_spinlock, flags);

	while (tempvoice != NULL) {

		if (voices->flags & VOICEMGR_FLAGS_PLAYBACK) {
			int voicenum = tempvoice->voicenum;

			sblive_writeptr(card, IFATN, voicenum, 0xffff);
			sblive_writeptr(card, IP, voicenum, 0);
			sblive_writeptr(card, VTFT, voicenum, 0xffff);
			sblive_writeptr(card, PTRX_PITCHTARGET, voicenum, 0);
		}

		tempvoice->voicestate = VOICEMGR_STATE_IDLE;
		tempvoice = tempvoice->linked_voice;
	}

	tempvoice = voices;

	spin_unlock_irqrestore(&sblive_spinlock, flags);

	return;
}

void sblive_voice_setcontrol(struct emu_voice *voices, struct voice_cntlset *setting, u32 numparam)
{
	struct emu10k1_card *card = voices->card;
	struct emu_voice *currvoice = voices;
	int count;

	for (count = 0; count < numparam; count++)
		sblive_writeptr(card, setting[count].paramID, currvoice->voicenum, setting[count].value);

	return;
}

void sblive_voice_getcontrol(struct emu_voice *voices, u32 controlid, u32 * value)
{
	struct emu10k1_card *card = voices->card;

	*value = sblive_readptr(card, controlid, voices->voicenum);

	return;
}
